// Copyright 1998-2017 Epic Games, Inc. All Rights Reserved.

#include "BiscuitCharacter.h"
#include "UObject/ConstructorHelpers.h"
#include "Camera/CameraComponent.h"
#include "Components/DecalComponent.h"
#include "Components/CapsuleComponent.h"
#include "GameFramework/CharacterMovementComponent.h"
#include "GameFramework/PlayerController.h"
#include "GameFramework/SpringArmComponent.h"
#include "Kismet/HeadMountedDisplayFunctionLibrary.h"
#include "Materials/Material.h"

ABiscuitCharacter::ABiscuitCharacter()
{
	// Set size for player capsule
	GetCapsuleComponent()->InitCapsuleSize(42.f, 96.0f);

	// Don't rotate character to camera direction
	bUseControllerRotationPitch = false;
	bUseControllerRotationYaw = false;
	bUseControllerRotationRoll = false;

	// Configure character movement
	GetCharacterMovement()->bOrientRotationToMovement = true; // Rotate character to moving direction
	GetCharacterMovement()->RotationRate = FRotator(0.f, 640.f, 0.f);
	GetCharacterMovement()->bConstrainToPlane = true;
	GetCharacterMovement()->bSnapToPlaneAtStart = true;

	// Create a camera boom...
	CameraBoom = CreateDefaultSubobject<USpringArmComponent>(TEXT("CameraBoom"));
	CameraBoom->SetupAttachment(RootComponent);
	CameraBoom->bAbsoluteRotation = true; // Don't want arm to rotate when character does
	CameraBoom->TargetArmLength = 800.f;
	CameraBoom->RelativeRotation = FRotator(-60.f, 0.f, 0.f);
	CameraBoom->bDoCollisionTest = false; // Don't want to pull camera in when it collides with level

	// Create a camera...
	TopDownCameraComponent = CreateDefaultSubobject<UCameraComponent>(TEXT("TopDownCamera"));
	TopDownCameraComponent->SetupAttachment(CameraBoom, USpringArmComponent::SocketName);
	TopDownCameraComponent->bUsePawnControlRotation = false; // Camera does not rotate relative to arm

	// Create a decal in the world to show the cursor's location
	CursorToWorld = CreateDefaultSubobject<UDecalComponent>("CursorToWorld");
	CursorToWorld->SetupAttachment(RootComponent);
	static ConstructorHelpers::FObjectFinder<UMaterial> DecalMaterialAsset(TEXT("Material'/Game/TopDownCPP/Blueprints/M_Cursor_Decal.M_Cursor_Decal'"));
	if (DecalMaterialAsset.Succeeded())
	{
		CursorToWorld->SetDecalMaterial(DecalMaterialAsset.Object);
	}
	CursorToWorld->DecalSize = FVector(16.0f, 32.0f, 32.0f);
	CursorToWorld->SetRelativeRotation(FRotator(90.0f, 0.0f, 0.0f).Quaternion());

	// Activate ticking in order to update the cursor every frame.
	PrimaryActorTick.bCanEverTick = true;
	PrimaryActorTick.bStartWithTickEnabled = true;
}

void ABiscuitCharacter::Tick(float DeltaSeconds)
{
    Super::Tick(DeltaSeconds);

	if (CursorToWorld != nullptr)
	{
		if (UHeadMountedDisplayFunctionLibrary::IsHeadMountedDisplayEnabled())
		{
			if (UWorld* World = GetWorld())
			{
				FHitResult HitResult;
				FCollisionQueryParams Params;
				FVector StartLocation = TopDownCameraComponent->GetComponentLocation();
				FVector EndLocation = TopDownCameraComponent->GetComponentRotation().Vector() * 2000.0f;
				Params.AddIgnoredActor(this);
				World->LineTraceSingleByChannel(HitResult, StartLocation, EndLocation, ECC_Visibility, Params);
				FQuat SurfaceRotation = HitResult.ImpactNormal.ToOrientationRotator().Quaternion();
				CursorToWorld->SetWorldLocationAndRotation(HitResult.Location, SurfaceRotation);
			}
		}
		else if (APlayerController* PC = Cast<APlayerController>(GetController()))
		{
			FHitResult TraceHitResult;
			PC->GetHitResultUnderCursor(ECC_Visibility, true, TraceHitResult);
			FVector CursorFV = TraceHitResult.ImpactNormal;
			FRotator CursorR = CursorFV.Rotation();
			CursorToWorld->SetWorldLocation(TraceHitResult.Location);
			CursorToWorld->SetWorldRotation(CursorR);
		}
	}
}

void ABiscuitCharacter::BeginPlay() {
	// Call the base class  
	Super::BeginPlay();

	// Check the selected UI class is not NULL
	if (SfsChatUIClass == NULL)
	{
		return;
	}

	// Check if the widget is already created
	if (SfsChatWidget != NULL)
	{
		return;
	}

	// Create Widget
	SfsChatWidget = CreateWidget<USfsChatWidget>(GetWorld(), SfsChatUIClass);
	if (!SfsChatWidget)
	{
		return;
	}

	// Get a reference to this character class to SfsChatWidget 
	// In this way SfsChatWidget could access SmartFox Server 
	// This is a shortcut for demo: please use best approach in a real product (interface or share smartfox instance or other)
	SfsChatWidget->SetBiscuitCharacter(this);

	// Add it to the viewport so the Construct() method in the UUserWidget:: is run.
	SfsChatWidget->AddToViewport();

	// Set it to visible 
	SfsChatWidget->SetVisibility(ESlateVisibility::Visible);

	// Set it to login mode
	SfsChatWidget->SetLoginMode();
}

void ABiscuitCharacter::SmartFoxInitialize()
{
	// Initialize Smart Fox
	m_ptrSmartFox = boost::shared_ptr<Sfs2X::SmartFox>(new Sfs2X::SmartFox(true));
	m_ptrSmartFox->ThreadSafeMode(false);

	// Add event listeners
	m_ptrSmartFox->AddEventListener(SFSEvent::CONNECTION, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxConnection, (unsigned long long)this)));
	m_ptrSmartFox->AddEventListener(SFSEvent::CONNECTION_LOST, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxConnectionLost, (unsigned long long)this)));
	m_ptrSmartFox->AddEventListener(SFSEvent::LOGIN, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxLogin, (unsigned long long)this)));
	m_ptrSmartFox->AddEventListener(SFSEvent::LOGIN_ERROR, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxLoginError, (unsigned long long)this)));
	m_ptrSmartFox->AddEventListener(SFSEvent::LOGOUT, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxLogout, (unsigned long long)this)));
	m_ptrSmartFox->AddEventListener(BitSwarmEvent::DISCONNECT, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxDisconnection, (unsigned long long)this)));
	m_ptrSmartFox->AddEventListener(SFSEvent::CONNECTION_RETRY, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxConnectionRetry, (unsigned long long)this)));
	m_ptrSmartFox->AddEventListener(SFSEvent::CONNECTION_RESUME, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxConnectionResume, (unsigned long long)this)));
	m_ptrSmartFox->AddEventListener(SFSEvent::ROOM_JOIN, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxRoomJoin, (unsigned long long)this)));
	m_ptrSmartFox->AddEventListener(SFSEvent::ROOM_JOIN_ERROR, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxRoomJoinError, (unsigned long long)this)));
	m_ptrSmartFox->AddEventListener(SFSEvent::PUBLIC_MESSAGE, boost::shared_ptr<EventListenerDelegate>(new EventListenerDelegate(ABiscuitCharacter::OnSmartFoxPublicMessage, (unsigned long long)this)));

	m_ptrSmartFox->LoadConfig("F:/FlyingCookie/Biscuit/sfs-config.xml", true);

	// Wait for connection establishment
	m_bSmartFoxConnectionEstablished = false;
	::DWORD dwRc = ::WaitForSingleObject(m_hSmartFoxConnectionEstablished, SMARTFOX_MAXIMUMWAITFORCONNECTIONESTABLISHMENT);

	if (dwRc == WAIT_TIMEOUT)
	{
		// Timeout
		return;
	}

	// Set flag to preserve the status of connection with SmartFox Server
	m_bSmartFoxConnectionEstablished = true;
}

void ABiscuitCharacter::SmartFoxLogin(string username)
{
	m_ptrSmartFoxLoginRequest = boost::shared_ptr<IRequest>(new LoginRequest(username, "", "BasicExamples"));
	m_ptrSmartFox->Send(m_ptrSmartFoxLoginRequest);
}

void ABiscuitCharacter::SmartFoxJoinRoom(string roomName)
{
	m_ptrSmartFoxJoinRoomRequest = boost::shared_ptr<IRequest>(new JoinRoomRequest(roomName));
	m_ptrSmartFox->Send(m_ptrSmartFoxJoinRoomRequest);
}

void ABiscuitCharacter::SmartFoxSendChatMessage(string message)
{
	//m_ptrSmartFoxPublicMessageRequest = boost::shared_ptr<IRequest>(new PublicMessageRequest(message));
	//m_ptrSmartFox->Send(m_ptrSmartFoxPublicMessageRequest);
}

bool ABiscuitCharacter::IsSmartFoxServerConnected()
{
	return m_bSmartFoxConnectionEstablished;
}

void ABiscuitCharacter::OnSmartFoxConnection(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}

	boost::shared_ptr<map<string, boost::shared_ptr<void>>> ptrEventParams = ptrEvent->Params();
	boost::shared_ptr<void> ptrEventParamValueSuccess = (*ptrEventParams)["success"];
	boost::shared_ptr<bool> ptrErrorMessage = ((boost::static_pointer_cast<bool>))(ptrEventParamValueSuccess);

	if (*ptrErrorMessage == true)
	{
		// Connection with SmartFox Server has been established
		// Set internal event to notify the connection establishment
		SetEvent(ptrABiscuitCharacter->m_hSmartFoxConnectionEstablished);
	}
}

void ABiscuitCharacter::OnSmartFoxConnectionLost(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}

	// No action when reconnection will be attempted
	if (ptrABiscuitCharacter->m_ptrSmartFox->GetReconnectionSeconds() > 0)
	{
		return;
	}
}

void ABiscuitCharacter::OnSmartFoxConnectionRetry(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}
}

void ABiscuitCharacter::OnSmartFoxConnectionResume(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}
}

void ABiscuitCharacter::OnSmartFoxDisconnection(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}

	// Set widget to login mode
	ptrABiscuitCharacter->SfsChatWidget->SetLoginMode();
}

void ABiscuitCharacter::OnSmartFoxLogin(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}

	// Join default room
	ptrABiscuitCharacter->SmartFoxJoinRoom("The Lobby");
}

void ABiscuitCharacter::OnSmartFoxLoginError(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}

	// Get the error message parameter of the event
	// Get the error code parameter of the event
	boost::shared_ptr<map<string, boost::shared_ptr<void>>> ptrEventParams = ptrEvent->Params();
	boost::shared_ptr<void> ptrEventParamValueErrorMessage = (*ptrEventParams)["errorMessage"];
	boost::shared_ptr<string> ptrErrorMessage = ((boost::static_pointer_cast<string>))(ptrEventParamValueErrorMessage);

	boost::shared_ptr<void> ptrEventParamValueErrorCode = (*ptrEventParams)["errorCode"];
	boost::shared_ptr<short int> ptrErrorCode = ((boost::static_pointer_cast<short int>))(ptrEventParamValueErrorCode);

	// Show error message
	UE_LOG(LogTemp, Warning, TEXT("Unable to login"));
}

void ABiscuitCharacter::OnSmartFoxRoomJoin(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}

	// Set widget to chat mode
	ptrABiscuitCharacter->SfsChatWidget->SetChatMode();
}

void ABiscuitCharacter::OnSmartFoxRoomJoinError(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}
}

void ABiscuitCharacter::OnSmartFoxPublicMessage(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}

	// Get the sender parameter of the event
	// Get the message parameter of the event
	boost::shared_ptr<map<string, boost::shared_ptr<void>>> ptrEventParams = ptrEvent->Params();
	boost::shared_ptr<void> ptrEventParamValueSender = (*ptrEventParams)["sender"];
	boost::shared_ptr<User> ptrNotifiedUser = ((boost::static_pointer_cast<User>))(ptrEventParamValueSender);

	boost::shared_ptr<void> ptrEventParamValueMessage = (*ptrEventParams)["message"];
	boost::shared_ptr<string> ptrNotifiedMessage = ((boost::static_pointer_cast<string>))(ptrEventParamValueMessage);

	// Show message
	ptrABiscuitCharacter->SfsChatWidget->ShowChatMessage(ptrNotifiedUser->Name()->c_str(), ptrNotifiedMessage->c_str());
}

void ABiscuitCharacter::OnSmartFoxLogout(unsigned long long ptrContext, boost::shared_ptr<BaseEvent> ptrEvent)
{
	// Get the pointer to ABiscuitCharacter class
	ABiscuitCharacter* ptrABiscuitCharacter = (ABiscuitCharacter*)ptrContext;

	if (ptrABiscuitCharacter == NULL)
	{
		return;
	}
}
